from matrizes import matrizes
from instancias import instancias
from itertools import combinations
import time
import colorama
colorama.init()

# Define códigos de escape ANSI para cores
cor_vermelha = colorama.Fore.RED #"\033[91m"
cor_verde = colorama.Fore.GREEN # "\033[92m"
cor_amarela = "\033[93m"
cor_azul = "\033[94m"
cor_reset = colorama.Fore.WHITE # "\033[0m"  # Reset para a cor padrão

# custa 4 : 3x3
# custa 5 : 5x5
# custa 9 : 7x7
# custa 17 : 9x9
#
class Estado:
    def __init__(self, submatriz, total, top_left, solucao):
        self.submatriz = submatriz
        self.total = total
        self.top_left = top_left
        self.children = []
        self.solucao = solucao


def criar_matriz_solucao(matriz, node, primeiravez=True):
    matriz_solucao = [row[:] for row in matriz]
    # Dimensões da matriz original
    linhas = len(matriz)
    colunas = len(matriz[0]) if matriz else 0
    custo = 0


    if primeiravez:
        # Criar uma nova matriz com zeros
        matriz_solucao = [[cor_reset] * colunas for _ in range(linhas)]

    if node is not None:
        if node.solucao:
            if hasattr(node, 'submatriz'):

                raio = (len(node.submatriz) -1 ) // 2
                if raio == 1:
                    custo = 4
                else:
                    if raio == 2:
                        custo = 5
                    else:
                        if raio == 3:
                            custo = 9
                        else:
                            if raio == 4:
                                custo = 17
                centro_linha = node.top_left[0] + raio
                centro_coluna = node.top_left[1] + raio
                dimensao_submatriz = len(node.submatriz)
                inicio_linha = node.top_left[0]
                inicio_coluna = node.top_left[1]
                for linha in range(dimensao_submatriz):
                    for coluna in range(dimensao_submatriz):
                        if centro_linha == (linha + inicio_linha)  and centro_coluna == (coluna + inicio_coluna ):
                            matriz_solucao[linha + inicio_linha][coluna + inicio_coluna] = cor_vermelha
                        else:
                            matriz_solucao[linha + inicio_linha][coluna + inicio_coluna] = cor_verde

            if hasattr(node, 'children'):
                for child in node.children:
                    matriz_solucao, custo_a = criar_matriz_solucao(matriz_solucao, child, False)
                    custo += custo_a

    return matriz_solucao, custo

def imprimir_matriz_com_moldura(matriz):
    # Adiciona uma linha de moldura no topo
    print("-" * (len(matriz[0]) * 4 + 2))

    # Itera sobre cada linha da matriz
    for linha in matriz:
        # Adiciona um caractere de moldura à esquerda
        print("|", end="")
        # Itera sobre cada elemento da linha e imprime o valor
        for elemento in linha:

            texto_formatado = "{:^4}".format(elemento)
            print(cor_reset +  texto_formatado, end="")
        # Adiciona um caractere de moldura à direita
        print("|")

    # Adiciona uma linha de moldura na parte inferior
    print("-" * (len(matriz[0]) * 4 + 2))

def imprimir_matriz_com_moldura_e_solucao(matriz, matriz_solucao, alvo, verba, custo):

    #node tera os elementos da solucao
    #quando colocar um elemento da matriz verifica se é
    #parte da solucao ou nao

    # Adiciona uma linha de moldura no topo
    print(cor_reset +"-" * (len(matriz[0]) * 4 + 2))
    nr_linha = 0
    nr_coluna = 0
    protegidos = 0
    delegacias = 0
    # Itera sobre cada linha da matriz
    for linha in matriz:
        nr_coluna = 0
        # Adiciona um caractere de moldura à esquerda
        print(cor_reset +"|", end="")
        # Itera sobre cada elemento da linha e imprime o valor
        for elemento in linha:

            texto_formatado = "{:^4}".format(elemento)
            #verifica linha,coluna
            # a funcao vai returnar a cor a colocar

            cor_texto = matriz_solucao[nr_linha][nr_coluna]
            if cor_texto == cor_vermelha:
                delegacias += 1
            if cor_texto == cor_vermelha or cor_texto == cor_verde:
                protegidos += elemento
            #cor_texto = check_solucao(linha,coluna,no)
            print(cor_texto + texto_formatado, end="")
            nr_coluna += 1
        # Adiciona um caractere de moldura à direita
        print(cor_reset +"|")
        nr_linha += 1
    # Adiciona uma linha de moldura na parte inferior
    print(cor_reset + "-" * (len(matriz[0]) * 4 + 2))

    print(cor_reset + "Disponivel: " + str(verba - custo) + " [ " + str(delegacias) + " delegacias colocadas]")
    print(cor_reset + "Protegidos: " + str(protegidos) + " / " + str(alvo) )
    print(cor_reset + "Solucao!")

#imprimir_matriz_com_moldura(matrizes[0])

#imprimir_matriz_com_moldura(matrizes[8])

def combinacoes_delegacias(custo_delegacias, valor_alvo):
    # Inicializa um conjunto para armazenar as combinações únicas
    combinacoes_unicas = set()

    # Função auxiliar recursiva para encontrar as combinações
    def encontrar_combinacoes(atual_combinacao, soma_combinacao, indice_custo):
        # Verifica se a soma é menor ou igual ao valor alvo
        if soma_combinacao <= valor_alvo:
            combinacoes_unicas.add(tuple(atual_combinacao))
        # Caso base: se a soma for maior que o valor alvo ou todas as moedas foram consideradas
        if soma_combinacao > valor_alvo or indice_custo == len(custo_delegacias):
            return

        # Caso recursivo: inclui o custo atual na combinação e chama recursivamente
        encontrar_combinacoes(atual_combinacao + [custo_delegacias[indice_custo]], soma_combinacao + custo_delegacias[indice_custo],
                              indice_custo)
        # Caso recursivo: não inclui o custo atual na combinação e chama recursivamente
        encontrar_combinacoes(atual_combinacao, soma_combinacao, indice_custo + 1)

    # Chama a função auxiliar com a lista vazia para começar, a soma inicial 0 e o índice inicial do custo
    encontrar_combinacoes([], 0, 0)

    # Converte as combinações únicas de volta para listas
    combinacoes_unicas = [list(comb) for comb in combinacoes_unicas]

    # Remove valores vazios
    combinacoes_unicas = [comb for comb in combinacoes_unicas if comb]

    # Ordena a lista de combinações pelo somatório de cada combinação
    #combinacoes_unicas.sort(key=sum)
    # Ordena a lista de combinações pelo comprimento de cada combinação
    # As menores sao mais rapidas de testar
    combinacoes_unicas.sort(key=len)
    return combinacoes_unicas

def troca_custo_por_matriz(lista):
    return list(map(lambda x: 3 if x == 4 else (7 if x == 9 else (9 if x == 17 else x)), lista))


def depth_first_search(matriz, dimensions, nivel, alvo):
    n_expansoes = 0
    n_geracoes = 0
    if len(dimensions)!=0:

        #, exclude_top_left, dimension):
        n = len(matriz)
        m = len(matriz[0])
        root_nodes = []

        # fazer uma copia da matriz atual
        original_matrix_copy = [row[:] for row in matriz]
        # fazer uma copia da matriz atual para trabalhar
        working_matriz = [row[:] for row in matriz]
        encontrou = False
        # Processa a dimensao da matriz
        i = dimensions[nivel]
        for r in range(n - i + 1):
            for c in range(m - i + 1):
                submatriz = [row[c:c + i] for row in matriz[r:r + i]]
                #verificar se tem algum elemento do pai
                #testava negativo para não haver sobreposicao
                #nunca há
                has_negative_one = any(-1 in row for row in submatriz)
                if (not has_negative_one):
                    total = sum(sum(row) for row in submatriz)

                    top_left = (r, c)
                    node = Estado(submatriz, total, top_left, False)
                    n_geracoes += 1
                    if total >= alvo:
                        encontrou = True
                        node.solucao = True
                    # colocar os valores da submatriz a zero para nao serem
                    # contabilizados nos outrso niveis
                    for x in range(r, r + i):
                        for y in range(c, c + i):
                            # aqui se colocado a -1 passa a não permitir sobreposição
                            working_matriz[x][y] = 0

                    #Agora posso criar o segundo nivel?
                    if ((len(dimensions)-1) != nivel) and (encontrou == False):
                        nivel += 1
                        n_expansoes += 1
                        #trata a expansao
                        node.children, a_expansoes, a_geracoes, encontrou = depth_first_search(working_matriz, dimensions, nivel, alvo - total)
                        n_geracoes += a_geracoes
                        n_expansoes += a_expansoes
                        node.solucao = encontrou
                        nivel -= 1
                    if node.solucao:
                        #adiciona só nos parte da solucao
                        root_nodes.append(node)
                    if encontrou:
                        return root_nodes, n_expansoes, n_geracoes, True
                working_matriz = [row[:] for row in matriz]

        return root_nodes, n_expansoes, n_geracoes, encontrou

def print_tree(node, depth=0):
    if node is not None:
        if node.solucao:
            if hasattr(node, 'submatriz'):
                print("  " * depth + f"Populacao:")
                for row in node.submatriz:
                    print("  " * depth + f"{row}")
                raio = (len(node.submatriz) -1 ) // 2
                centro_linha = node.top_left[0] + raio + 1
                centro_coluna = node.top_left[1] + raio + 1
                print("  " * depth + f"Total: {node.total}, Top-left: {node.top_left} , Center: {centro_linha}, {centro_coluna}, Raio: {raio} , Solucao: {node.solucao}")
            if hasattr(node, 'children'):
                for child in node.children:
                    print_tree(child, depth + 1)

def print_dados(nr_instancia, alvo, dimensoes, expansoes, geracoes, tempo_processamento, estado):


    # Imprimindo dados
    linha_formatada = "| {:<10} | {:<10} | {:<15} | {:<10} | {:<10} | {:<20.2f} | {:<30} |".format(
        nr_instancia, alvo, str(dimensoes),  expansoes, geracoes, tempo_processamento, estado
    )
    print(linha_formatada)

    #print("-" * len(cabecalho))

def nodes_sol(node):
    top_lefts = []
    try:
        if node.children:
            for ch in node.children:
                aux=get_solucao(ch)
                top_lefts.append(aux)
        if node.solucao:
            top_lefts.append(str(node.top_left) + ' ' + str(len(node.submatriz)))
        return top_lefts
    except:
        top_lefts.append(str(node.top_left) + ' ' + str(len(node.submatriz)))
        return top_lefts

def get_solucao(root_nodes):
    top_lefts = []
    try:
        # root_nodes é uma lista de Estado objects
        for node in root_nodes:
            if node.children:
                for ch in node.children:
                    aux=nodes_sol(ch)
                    top_lefts.append(aux)
            if node.solucao:
                top_lefts.append(str(node.top_left) + ' ' + str(len(node.submatriz)))
        return top_lefts
    except:
        top_lefts.append(str(root_nodes.top_left) + ' ' + str(len(root_nodes.submatriz)))
        return top_lefts

# Custo das Delegacias disponíveis
custo_delegacias = [4, 5, 9, 17]

#delegacias = combinacoes_delegacias(custo_delegacias, 12)

def processar_todas():
    #expansao - quando se cria uma nova posicao, um novo nivel
    #geracao - conjunto das diferentes combinacoes de posicoes possiveis dentro da matriz - todas as combinacoes testadas
    for i in range(10):
        parametros = instancias[i] # id - n - m - verba - protegerA - protegerB
        mapa = matrizes[i]
        verba = parametros [3]
        #obter a combinacao de delegacias - custo
        delegacias = combinacoes_delegacias(custo_delegacias, verba)

        alvos =  []
        alvos.append(parametros[4]) # protegerA
        alvos.append(parametros[5]) # protegerB
        # Imprimindo cabeçalho
        cabecalho = "| {:<10} | {:<10} | {:<15} | {:<10} | {:<10} | {:<20} | {:<30} |".format(
            "Instância", "Alvo", "Dimensões", "Expansões", "Gerações", "Tempo Process.(ms)", "Encontrou"
        )

        for alvo in alvos:
            print("-" * len(cabecalho))
            print(cabecalho)
            print("-" * len(cabecalho))
            tempo_inicio_alvo = time.time() * 1000
            total_expansoes = 0
            total_geracoes = 0
            encontrou = False
            for custo in delegacias:

                # Substitui os valores na lista
                copia_dimensoes = troca_custo_por_matriz(custo)
                #print(str(copia_dimensoes))

                nivel = 0
                encontrou = False
                start_time = time.time() * 1000

                root_nodes, expansoes, geracoes, encontrou = depth_first_search(mapa, copia_dimensoes, nivel, alvo)
                processing_time = (time.time() * 1000 - start_time)
                total_expansoes += expansoes
                total_geracoes += geracoes
                #print(f"instancia {i} alvo: {alvo} {copia_dimensoes}  expansoes: {expansoes} geracoes: {geracoes} processing_time: {processing_time:.6f} seconds Encontrou: {encontrou}")
                print_dados(i+1, alvo, copia_dimensoes, expansoes, geracoes, processing_time, encontrou)
                if encontrou:
                    print("-" * len(cabecalho))
                    processing_time = (time.time() * 1000 - tempo_inicio_alvo)
                    aaa = get_solucao(root_nodes)
                    print_dados('Fim: ' + str(i + 1), alvo, '', total_expansoes, total_geracoes, processing_time, str(aaa))
                    print("-" * len(cabecalho))
                    print_tree(root_nodes[0], 0)
                    print("\n")
                    m_solucao, custo = criar_matriz_solucao(mapa, root_nodes[0], True)
                    imprimir_matriz_com_moldura_e_solucao(mapa, m_solucao, alvo, verba, custo)
                    break

            #if not encontrou:
             #   print_dados(i + 1, alvo, copia_dimensoes, expansoes, geracoes, processing_time, encontrou)
        print("-" * len(cabecalho))
        processing_time = (time.time() * 1000 - tempo_inicio_alvo)
        print_dados('Fim: ' + str(i + 1), alvo, '', total_expansoes, total_geracoes, processing_time, '')
        print("-" * len(cabecalho))
        print("Sem Solução !")

def processar_instancia( id_instancia):
    #expansao - quando se cria uma nova posicao, um novo nivel
    #geracao - conjunto das diferentes combinacoes de posicoes possiveis dentro da matriz - todas as combinacoes testadas
    i = id_instancia - 1
    parametros = instancias[i] # id - n - m - verba - protegerA - protegerB
    mapa = matrizes[i]
    verba = parametros [3]
    #obter a combinacao de delegacias - custo
    delegacias = combinacoes_delegacias(custo_delegacias, verba)

    alvos =  []
    alvos.append(parametros[4]) # protegerA
    alvos.append(parametros[5]) # protegerB
    # Imprimindo cabeçalho
    cabecalho = "| {:<10} | {:<10} | {:<15} | {:<10} | {:<10} | {:<20} | {:<30} |".format(
        "Instância", "Alvo", "Dimensões", "Expansões", "Gerações", "Tempo Process.(ms)", "Encontrou"
    )
    #print("-" * len(cabecalho))
    #print(cabecalho)
    #print("-" * len(cabecalho))
    for alvo in alvos:
        print("-" * len(cabecalho))
        print(cabecalho)
        print("-" * len(cabecalho))
        encontrou = False
        tempo_inicio_alvo = time.time() * 1000
        total_expansoes = 0
        total_geracoes = 0
        for custo in delegacias:

            # Substitui os valores na lista
            copia_dimensoes = troca_custo_por_matriz(custo)

            nivel = 0
            encontrou = False
            start_time = time.time() * 1000
            root_nodes, expansoes, geracoes, encontrou = depth_first_search(mapa, copia_dimensoes, nivel, alvo)
            processing_time = (time.time() * 1000 - start_time)
            total_expansoes += expansoes
            total_geracoes += geracoes
            #print(f"instancia {i} alvo: {alvo} {copia_dimensoes}  expansoes: {expansoes} geracoes: {geracoes} processing_time: {processing_time:.6f} seconds Encontrou: {encontrou}")
            print_dados(i+1, alvo, copia_dimensoes, expansoes, geracoes, processing_time, encontrou)
            if encontrou:
                print("-" * len(cabecalho))
                processing_time = (time.time() * 1000 - tempo_inicio_alvo)
                aaa = get_solucao(root_nodes)
                print_dados('Fim: ' + str(i + 1), alvo, '', total_expansoes, total_geracoes, processing_time, str(aaa))
                print("-" * len(cabecalho))
                print_tree(root_nodes[0], 0)
                print("\n")
                m_solucao, custo = criar_matriz_solucao(mapa, root_nodes[0], True)
                imprimir_matriz_com_moldura_e_solucao(mapa, m_solucao, alvo, verba, custo)
                break
        #if not encontrou:
         #   print_dados(i + 1, alvo, copia_dimensoes, expansoes, geracoes, processing_time, encontrou)
    print("-" * len(cabecalho))
    processing_time = (time.time() * 1000 - tempo_inicio_alvo)
    print_dados('Fim: ' + str(i + 1), alvo, '', total_expansoes, total_geracoes, processing_time, '')
    print("-" * len(cabecalho))
    print("Sem Solução !" )
def escolheu_instancia():
    instancia_num = 0
    while instancia_num == 0:
        resposta = input("Digite o número da instância (entre 1 e 10): ")
        if not resposta.isdigit():
            return
        else:
            instancia_num = int(resposta)
            if int(instancia_num) < 1 or int(instancia_num) > 10:
                print("Número de instância inválido.")
                instancia_num = 0
                return

    if int(instancia_num) >= 1 and int(instancia_num)<=10:
        processar_instancia(int(instancia_num))


while True:
    print("\nMenu:")
    print("1. Processar todas")
    print("2. Processar instância")
    print("0. Sair")
    opcao = input("Escolha uma opção: ")

    if opcao == "1":
        processar_todas()
    elif opcao == "2":
        escolheu_instancia()
    elif opcao == "0":
        print("Saindo do programa...")
        break
    else:
        print("Opção inválida. Por favor, escolha uma opção válida.")